@@ -145,7 +145,7 @@ char_is_function_name_char(int b)
 static bool
 char_is_function_argument_char(int b)
 {
-  if (char_is_delimiter(b) || b == '(' || b == ')')
+  if (b == '(' || b == ')')
     return false;
   return true;
 }
@@ -1365,7 +1365,7 @@ public:
   ///
   /// A string is just a contiguous set of characters that test
   /// positive when passed to
-  /// read_context::char_is_property_name_char().
+  /// read_context::char_is_property_value_char().
   ///
   /// Note that all escaped characters are suitable to be in a string.
   ///
