@@ -1,7 +1,7 @@
 // SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
 // -*- Mode: C++ -*-
 //
-// Copyright (C) 2013-2022 Red Hat, Inc.
+// Copyright (C) 2013-2023 Red Hat, Inc.
 //
 // Author: Sinny Kumari
 
@@ -149,6 +149,19 @@ static InOutSpec in_out_specs[] =
     "data/test-diff-pkg/dirpkg-3-report-2.txt",
     "output/test-diff-pkg/dirpkg-3-report-2.txt"
   },
+  {
+    "data/test-diff-pkg/symlink-dir-test1/dir1/targets",
+    "data/test-diff-pkg/symlink-dir-test1/dir2/targets",
+    "--no-default-suppression ",
+    "",
+    "",
+    "",
+    "",
+    "",
+    "data/test-diff-pkg/symlink-dir-test1-report0.txt",
+    "output/test-diff-pkg/symlink-dir-test1-report0.txt"
+  },
+#ifdef WITH_SYMLINKS_KEPT_IN_DIST
   {
     "data/test-diff-pkg/symlink-dir-test1/dir1/symlinks",
     "data/test-diff-pkg/symlink-dir-test1/dir2/symlinks",
@@ -158,9 +171,10 @@ static InOutSpec in_out_specs[] =
     "",
     "",
     "",
-    "data/test-diff-pkg/symlink-dir-test1-report0.txt ",
-    "output/test-diff-pkg/symlink-dir-test1-report0.txt "
+    "data/test-diff-pkg/symlink-dir-test1-report1.txt",
+    "output/test-diff-pkg/symlink-dir-test1-report1.txt"
   },
+#endif
 #if WITH_TAR
   {
     "data/test-diff-pkg/tarpkg-0-dir1.tar",
@@ -864,6 +878,19 @@ static InOutSpec in_out_specs[] =
     "data/test-diff-pkg-ctf/dirpkg-3-report-2.txt",
     "output/test-diff-pkg-ctf/dirpkg-3-report-2.txt"
   },
+  {
+    "data/test-diff-pkg-ctf/symlink-dir-test1/dir1/targets",
+    "data/test-diff-pkg-ctf/symlink-dir-test1/dir2/targets",
+    "--ctf --no-default-suppression ",
+    "",
+    "",
+    "",
+    "",
+    "",
+    "data/test-diff-pkg-ctf/symlink-dir-test1-report0.txt",
+    "output/test-diff-pkg-ctf/symlink-dir-test1-report0.txt"
+  },
+#ifdef WITH_SYMLINKS_KEPT_IN_DIST
   {
     "data/test-diff-pkg-ctf/symlink-dir-test1/dir1/symlinks",
     "data/test-diff-pkg-ctf/symlink-dir-test1/dir2/symlinks",
@@ -873,9 +900,10 @@ static InOutSpec in_out_specs[] =
     "",
     "",
     "",
-    "data/test-diff-pkg-ctf/symlink-dir-test1-report0.txt ",
-    "output/test-diff-pkg-ctf/symlink-dir-test1-report0.txt "
+    "data/test-diff-pkg-ctf/symlink-dir-test1-report1.txt",
+    "output/test-diff-pkg-ctf/symlink-dir-test1-report1.txt"
   },
+#endif
 #if WITH_TAR
   {
     "data/test-diff-pkg-ctf/tarpkg-0-dir1.tar",
